"""Manage annotations associated with Kubernetes resources

Different types of annotations can be attached to any Kubernetes resource or custom resource

The annotations are stored in various places - secrets, labels, annotations and configmaps
"""