PROVIDER_SUBMODULE='ckan-db'

MIGRATION_CRD_SINGULAR = 'dbmigration'
MIGRATION_CRD_PLURAL = 'dbmigrations'
MIGRATION_CRD_KIND = 'DbMigration'

MIGRATION_EVENT_STATUS_SUCCESS = 'success'
MIGRATION_EVENT_STATUS_FAILURE = 'failure'
