PROVIDER_SUBMODULE='apps'

APP_CRD_SINGULAR = 'app'
APP_CRD_PLURAL_SUFFIX = 'apps'
APP_CRD_KIND_SUFFIX = 'App'

APP_NAME_CRD_SINGULAR = 'appname'
APP_NAME_CRD_PLURAL_SUFFIX = 'appnames'
APP_NAME_CRD_KIND_SUFFIX = 'AppName'
