PROVIDER_SUBMODULE = 'cluster'

# Kubernetes namespace used to store operator resources
# this namespace should only keep objects created and managed by the operator
OPERATOR_NAMESPACE = 'ckan-cloud'

# Kubernetes configmap name in the operator namespace, used to store global operator configurations
OPERATOR_CONFIGMAP = 'operator-conf'
