PROVIDER_SUBMODULE='ckan'

INSTANCE_CRD_SINGULAR = 'ckaninstance'
INSTANCE_CRD_PLURAL_SUFFIX = 'ckaninstances'
INSTANCE_CRD_KIND_SUFFIX = 'CkanInstance'

INSTANCE_NAME_CRD_SINGULAR = 'ckaninstancename'
INSTANCE_NAME_CRD_PLURAL_SUFFIX = 'ckaninstancenames'
INSTANCE_NAME_CRD_KIND_SUFFIX = 'CkanInstanceName'
